import copy

import ghedt as dt
import numpy as np
import pygfunction as gt
import matplotlib.pyplot as plt


def square_and_near_square(lower: int,
                           upper: int,
                           B: float):
    if lower < 1 or upper < 1:
        raise ValueError('The lower and upper arguments must be positive'
                         'integer values.')
    if upper < lower:
        raise ValueError('The lower argument should be less than or equal to'
                         'the upper.')

    coordinates_domain = []

    for i in range(lower, upper+1):
        for j in range(2):
            coordinates = \
                dt.coordinates.rectangle(i, i+j, B, B)

            coordinates_domain.append(coordinates)

    return coordinates_domain


def rectangular(length_x, length_y, B_min, B_max, disp=False):
    # Make this work for the transpose
    if length_x >= length_y:
        length_1 = length_x
        length_2 = length_y
        transpose = False
    else:
        length_1 = length_y
        length_2 = length_x
        transpose = True

    rectangle_domain = []
    # find the maximum number of boreholes as a float
    n_1_max = (length_1 / B_min) + 1
    n_1_min = (length_1 / B_max) + 1

    N_min = int(np.ceil(n_1_min).tolist())
    N_max = int(np.floor(n_1_max).tolist())

    n_2_old = 1

    if disp:
        print(50 * '-')
        print('Rectangular Domain\nNx\tNy\tBx\tBy')

    iter = 0
    for N in range(N_min, N_max+1):
        # Check to see if we bracket
        B = length_1 / (N - 1)
        n_2 = int(np.floor((length_2 / B) + 1))

        if iter == 0:
            for i in range(1, N_min):
                r = dt.coordinates.rectangle(i, 1, B, B)
                if transpose:
                    r = dt.coordinates.transpose_coordinates(r)
                rectangle_domain.append(r)
            for j in range(1, n_2):
                r = dt.coordinates.rectangle(N_min, j, B, B)
                if transpose:
                    r = dt.coordinates.transpose_coordinates(r)
                rectangle_domain.append(r)

            iter += 1
        if n_2_old == n_2:
            pass
        else:
            r = dt.coordinates.rectangle(N, n_2, B, B)
            if disp:
                print('{}\t{}\t{}\t{}'.format(N, n_2, B, B))
            if transpose:
                r = dt.coordinates.transpose_coordinates(r)
            rectangle_domain.append(r)
            n_2_old = copy.deepcopy(n_2)

        N += 1

    return rectangle_domain


def bi_rectangular(
        length_x, length_y, B_min, B_max_x, B_max_y, transpose=False,
        disp=False):
    # Make this work for the transpose
    if length_x >= length_y:
        length_1 = length_x
        length_2 = length_y
        B_max_1 = B_max_x
        B_max_2 = B_max_y
    else:
        length_1 = length_y
        length_2 = length_x
        B_max_1 = B_max_y
        B_max_2 = B_max_x

    def func(B, length, n):
        _n = (length / B) + 1
        return n - _n

    bi_rectangle_domain = []
    # find the maximum number of boreholes as a float
    n_1_max = (length_1 / B_min) + 1
    n_1_min = (length_1 / B_max_1) + 1

    # if it is the first case in the domain, we want to step up from one
    # borehole, to a line, to adding the rows
    iter = 0

    N_min = int(np.ceil(n_1_min).tolist())
    N_max = int(np.floor(n_1_max).tolist())
    for n_1 in range(N_min, N_max + 1):

        n_2 = int(np.ceil((length_2 / B_max_2) + 1))
        b_2 = length_2 / (n_2 - 1)

        b_1 = length_1 / (n_1 - 1)

        if iter == 0:
            for i in range(1, n_1):
                coordinates = dt.coordinates.rectangle(i, 1, b_1, b_2)
                if transpose:
                    coordinates = \
                        dt.coordinates.transpose_coordinates(coordinates)
                bi_rectangle_domain.append(coordinates)
            for j in range(1, n_2):
                coordinates = dt.coordinates.rectangle(n_1, j, b_1, b_2)
                if transpose:
                    coordinates = \
                        dt.coordinates.transpose_coordinates(coordinates)
                bi_rectangle_domain.append(coordinates)

            iter += 1

        if disp:
            print('{0}x{1} with {2:.1f}x{3:.1f}'.format(n_1, n_2, b_1, b_2))

        coordinates = dt.coordinates.rectangle(n_1, n_2, b_1, b_2)
        if transpose:
            coordinates = \
                dt.coordinates.transpose_coordinates(coordinates)
        bi_rectangle_domain.append(coordinates)

        n_1 += 1

    return bi_rectangle_domain


def bi_rectangle_nested(length_x, length_y, B_min, B_max_x, B_max_y,
                        disp=False):
    # Make this work for the transpose
    if length_x >= length_y:
        length_1 = length_x
        length_2 = length_y
        B_max_1 = B_max_x
        B_max_2 = B_max_y
        transpose = False
    else:
        length_1 = length_y
        length_2 = length_x
        B_max_1 = B_max_y
        B_max_2 = B_max_x
        transpose = True

    # find the maximum number of boreholes as a float
    n_2_max = (length_2 / B_min) + 1
    n_2_min = (length_2 / B_max_2) + 1

    N_min = int(np.ceil(n_2_min).tolist())
    N_max = int(np.floor(n_2_max).tolist())

    bi_rectangle_nested_domain = []

    for n_2 in range(N_min, N_max + 1):
        b_2 = length_2 / (n_2 - 1)
        bi_rectangle_domain = dt.domains.bi_rectangular(
            length_1, length_2, B_min, B_max_1, b_2, transpose=transpose,
            disp=disp)
        bi_rectangle_nested_domain.append(bi_rectangle_domain)

    return bi_rectangle_nested_domain


def zoned_rectangle_domain(length_x, length_y, n_x, n_y, transpose=False):
    # Make this work for the transpose
    if length_x >= length_y:
        length_1 = length_x
        length_2 = length_y
        n_1 = n_x
        n_2 = n_y
    else:
        length_1 = length_y
        length_2 = length_x
        n_1 = n_y
        n_2 = n_x

    b_1 = length_1 / (n_1 - 1)
    b_2 = length_2 / (n_2 - 1)

    zoned_rectangle_domain = []

    n_i1 = 1
    n_i2 = 1

    z = dt.coordinates.zoned_rectangle(n_1, n_2, b_1, b_2, n_i1, n_i2)
    zoned_rectangle_domain.append(z)

    while n_i1 < (n_1 - 2) or n_i2 < (n_2 - 2):

        ratio = b_1 / b_2

        # general case where we can reduce in either direction
        # inner rectangular spacing
        bi_1 = (n_1 - 1) * b_1 / (n_i1 + 1)
        bi_2 = (n_2 - 1) * b_2 / (n_i2 + 1)
        # inner spacings for increasing each row
        bi_1_p1 = (n_1 - 1) * b_1 / (n_i1 + 2)
        bi_2_p1 = (n_2 - 1) * b_2 / (n_i2 + 2)

        ratio_1 = bi_1 / bi_2_p1
        ratio_2 = bi_2 / bi_1_p1

        # we only want to increase one at a time, and we want to increase
        # the one that will keep the inner rectangle furthest from the perimeter

        if ratio_1 > ratio:
            n_i1 += 1
        elif ratio_1 <= ratio:
            n_i2 += 1
        else:
            raise ValueError('This function should not have ever made it to '
                             'this point, there may be a problem with the '
                             'inputs.')
        z = dt.coordinates.zoned_rectangle(n_1, n_2, b_1, b_2, n_i1, n_i2)
        if transpose:
            z = dt.coordinates.transpose_coordinates(z)
        zoned_rectangle_domain.append(z)

    return zoned_rectangle_domain


def bi_rectangle_zoned_nested(length_x, length_y, B_min, B_max_x, B_max_y):
    # Make this work for the transpose
    if length_x >= length_y:
        length_1 = length_x
        length_2 = length_y
        B_max_1 = B_max_x
        B_max_2 = B_max_y
        transpose = False
    else:
        length_1 = length_y
        length_2 = length_x
        B_max_1 = B_max_y
        B_max_2 = B_max_x
        transpose = True

    # find the maximum number of boreholes as a float
    n_1_max = (length_1 / B_min) + 1
    n_1_min = (length_1 / B_max_1) + 1

    n_2_max = (length_2 / B_min) + 1
    n_2_min = (length_2 / B_max_2) + 1

    N_min_1 = int(np.ceil(n_1_min).tolist())
    N_max_1 = int(np.floor(n_1_max).tolist())

    N_min_2 = int(np.ceil(n_2_min).tolist())
    N_max_2 = int(np.floor(n_2_max).tolist())

    bi_rectangle_zoned_nested_domain = []

    n_1_values = list(range(N_min_1, N_max_1+1))
    n_2_values = list(range(N_min_2, N_max_2+1))

    j = 0  # pertains to n_1_values
    k = 0  # pertains to n_2_values
    l = 0

    for i in range(len(n_1_values) + len(n_2_values)-1):
        domain = []
        if l == 0:
            b_x = length_x / (N_min_1 - 1)
            b_y = length_y / (N_min_2 - 1)

            # go from one borehole to a line
            for l in range(1, N_min_1 + 1):
                r = dt.coordinates.rectangle(l, 1, b_x, b_y)
                if transpose:
                    r = dt.coordinates.transpose_coordinates(r)
                domain.append(r)

            # go from a line to an L
            for l in range(2, N_min_2 + 1):
                L = dt.coordinates.L_shape(N_min_1, l, b_x, b_y)
                if transpose:
                    L = dt.coordinates.transpose_coordinates(L)
                domain.append(L)

            # go from an L to a U
            for l in range(2, N_min_2 + 1):
                lop_u = \
                    dt.coordinates.lop_U(N_min_1, N_min_2, b_x, b_y, l)
                if transpose:
                    lop_u = dt.coordinates.transpose_coordinates(lop_u)
                domain.append(lop_u)

            # go from a U to an open
            for l in range(1, N_min_1 - 1):
                c = dt.coordinates.C_shape(N_min_1, N_min_2, b_x, b_y, l)
                if transpose:
                    c = dt.coordinates.transpose_coordinates(c)
                domain.append(c)

            l += 1

        if i % 2 == 0:
            bi_rectangle_zoned_domain = \
                zoned_rectangle_domain(length_1, length_2, n_1_values[j],
                                       n_2_values[k], transpose=transpose)
            domain.extend(bi_rectangle_zoned_domain)
            if j < len(n_1_values)-1:
                j += 1
            else:
                k += 1
        else:
            bi_rectangle_zoned_domain = \
                zoned_rectangle_domain(length_1, length_2, n_1_values[j],
                                       n_2_values[k], transpose=transpose)
            domain.extend(bi_rectangle_zoned_domain)
            if k < len(n_2_values)-1:
                k += 1
            else:
                j += 1

        bi_rectangle_zoned_nested_domain.append(domain)

    return bi_rectangle_zoned_nested_domain


def polygonal_land_constraint(property_boundary, B_min, B_max_x, B_max_y,
                              building_description=None):
    if building_description is None:
        building_description = []

    outer_rectangle = \
        dt.feature_recognition.determine_largest_rectangle(property_boundary)

    x, y = list(zip(*outer_rectangle))
    length = max(x)
    width = max(y)
    coordinates_domain_nested = \
            dt.domains.bi_rectangle_nested(length, width, B_min, B_max_x,
                                              B_max_y)

    coordinates_domain_nested_cutout = []

    for i in range(len(coordinates_domain_nested)):
        new_coordinates_domain = []
        for j in range(len(coordinates_domain_nested[i])):
            coordinates = coordinates_domain_nested[i][j]
            # Remove boreholes outside of property
            new_coordinates = dt.feature_recognition.remove_cutout(
                coordinates, boundary=property_boundary, remove_inside=False)
            # Remove boreholes inside of building
            if len(new_coordinates) == 0:
                continue
            new_coordinates = dt.feature_recognition.remove_cutout(
                new_coordinates, boundary=building_description,
                remove_inside=True, keep_contour=False)
            new_coordinates_domain.append(new_coordinates)
        coordinates_domain_nested_cutout.append(new_coordinates_domain)

    coordinates_domain_nested_cutout_reordered = []
    for i in range(len(coordinates_domain_nested_cutout)):
        domain = coordinates_domain_nested_cutout[i]
        domain_reordered = reorder_domain(domain)
        coordinates_domain_nested_cutout_reordered.append(domain_reordered)

    return coordinates_domain_nested_cutout_reordered


# The following functions are utility functions specific to domains.py
# ------------------------------------------------------------------------------
def verify_excess(domain):
    # Verify that the domain is unimodal
    unimodal = True
    delta_T_values = []
    for i in range(1, len(domain)):
        delta_T = domain[i] - domain[i-1]
        delta_T_values.append(delta_T)
        if delta_T > 0:
            unimodal = False

    return delta_T_values, unimodal


def reorder_domain(domain):
    # Reorder the domain so that the number of boreholes successively grow
    numbers = {}
    for i in range(len(domain)):
        numbers[i] = len(domain[i])

    sorted_values = sorted(numbers.values())

    reordered_domain = []

    for i in sorted_values:
        for j in numbers.keys():
            if numbers[j] == i:
                reordered_domain.append(domain[j])
                break

    return reordered_domain


def visualize_domain(domain, output_folder_name):
    import os
    if not os.path.exists(output_folder_name):
        os.makedirs(output_folder_name)

    for i in range(len(domain)):
        fig = gt.utilities._initialize_figure()
        ax = fig.add_subplot(111)

        coordinates = domain[i]
        x, y = list(zip(*coordinates))

        ax.scatter(x, y)

        ax.set_xlabel('x (m)')
        ax.set_ylabel('y (m)')

        fig.tight_layout()

        name = output_folder_name + '/' + str(i).zfill(3)

        fig.savefig(name)
        plt.close(fig)



import unittest
import ghedt.peak_load_analysis_tool as plat
import pygfunction as gt

"""
Test the single U-tube, double U-tube and coaxial borehole heat exchangers.
"""


class TestBHE(unittest.TestCase):

    def setUp(self) -> None:
        # Borehole dimensions
        self.H = 100.  # Borehole length (m)
        self.D = 2.  # Borehole buried depth (m)
        self.r_b = 150. / 1000. / 2.  # Borehole radius

        # Thermal conductivities
        self.k_p = 0.4  # Pipe thermal conductivity (W/m.K)
        k_s = 2.0  # Ground thermal conductivity (W/m.K)
        k_g = 1.0  # Grout thermal conductivity (W/m.K)

        # Volumetric heat capacities
        self.rhoCp_p = 1542. * 1000.  # Pipe volumetric heat capacity (J/K.m3)
        rhoCp_s = 2343.493 * 1000.  # Soil volumetric heat capacity (J/K.m3)
        rhoCp_g = 3901. * 1000.  # Grout volumetric heat capacity (J/K.m3)

        # Thermal properties
        # Soil
        ugt = 18.3  # Undisturbed ground temperature (degrees Celsius)
        self.soil = plat.media.Soil(k_s, rhoCp_s, ugt)
        # Grout
        self.grout = plat.media.Grout(k_g, rhoCp_g)

        # Fluid properties
        self.fluid = gt.media.Fluid('MEG', 0.)
        V_flow_borehole = 0.2  # Volumetric flow rate per borehole (L/s)
        # Total fluid mass flow rate per borehole (kg/s)
        self.m_flow_borehole = V_flow_borehole / 1000. * self.fluid.rho

    def test_single_u_tube(self):
        # Pipe dimensions
        r_out = 26.67 / 1000. / 2.  # Pipe outer radius (m)
        r_in = 21.6 / 1000. / 2.  # Pipe inner radius (m)
        s = 32.3 / 1000.  # Inner-tube to inner-tube Shank spacing (m)
        epsilon = 1.0e-6  # Pipe roughness (m)

        # Pipe positions
        # Single U-tube [(x_in, y_in), (x_out, y_out)]
        pos = plat.media.Pipe.place_pipes(s, r_out, 1)

        # Thermal properties
        # Pipe
        pipe = plat.media.Pipe(
            pos, r_in, r_out, s, epsilon, self.k_p, self.rhoCp_p)

        # Define a borehole
        borehole = gt.boreholes.Borehole(self.H, self.D, self.r_b, x=0., y=0.)

        single_u_tube = plat.borehole_heat_exchangers.SingleUTube(
            self.m_flow_borehole, self.fluid, borehole, pipe, self.grout, self.soil)

        Re = plat.borehole_heat_exchangers.compute_Reynolds(
            single_u_tube.m_flow_pipe, r_in, epsilon, self.fluid)

        # Reynolds number
        self.assertAlmostEqual(Re, 11667.361867903752, places=8)
        # Pipe resistance (K/(W/m))
        self.assertAlmostEqual(single_u_tube.R_p, 0.08389296717975074, places=8)
        # Convection coefficient (W/(m2.K))
        self.assertAlmostEqual(single_u_tube.h_f, 2522.2943235234375, places=8)
        # Convective resistance (K/(W/m))
        self.assertAlmostEqual(
            single_u_tube.R_fp, 0.08973549264054603, places=8)
        # Borehole thermal resistance ((m.K)/W)
        self.assertAlmostEqual(
            single_u_tube.compute_effective_borehole_resistance(),
            0.20730076327484714, places=8)

    def test_double_u_tube(self):
        # Pipe dimensions
        r_out = 26.67 / 1000. / 2.  # Pipe outer radius (m)
        r_in = 21.6 / 1000. / 2.  # Pipe inner radius (m)
        s = 32.3 / 1000.  # Inner-tube to inner-tube Shank spacing (m)
        epsilon = 1.0e-6  # Pipe roughness (m)

        # Pipe positions
        # Double U-tube [(x_in, y_in), (x_out, y_out), (x_in, y_in), (x_out, y_out)]
        pos = plat.media.Pipe.place_pipes(s, r_out, 2)

        # Thermal properties
        # Pipe
        pipe = plat.media.Pipe(pos, r_in, r_out, s, epsilon, self.k_p, self.rhoCp_p)

        # Define a borehole
        borehole = gt.boreholes.Borehole(self.H, self.D, self.r_b, x=0., y=0.)

        double_u_tube_parallel = plat.borehole_heat_exchangers.MultipleUTube(
            self.m_flow_borehole, self.fluid, borehole, pipe, self.grout,
            self.soil, config='parallel')

        # Intermediate variables
        Re = plat.borehole_heat_exchangers.compute_Reynolds(
            double_u_tube_parallel.m_flow_pipe, r_in, epsilon, self.fluid)

        # Reynolds number
        self.assertAlmostEqual(Re, 5833.680933951876, places=8)
        # Pipe resistance (K/(W/m))
        self.assertAlmostEqual(
            double_u_tube_parallel.R_p, 0.08389296717975074, places=8)
        # Convection coefficient (W/(m2.K))
        self.assertAlmostEqual(
            double_u_tube_parallel.h_f, 1292.3277737336607, places=8)
        # Convective resistance (K/(W/m))
        self.assertAlmostEqual(
            double_u_tube_parallel.R_fp, 0.09529608727383472, places=8)
        # Borehole thermal resistance ((m.K)/W)
        self.assertAlmostEqual(
            double_u_tube_parallel.compute_effective_borehole_resistance(),
            0.15970364062604958, places=8)

        double_u_tube_series = plat.borehole_heat_exchangers.MultipleUTube(
            self.m_flow_borehole, self.fluid, borehole, pipe, self.grout,
            self.soil, config='series')

        # Intermediate variables
        Re = plat.borehole_heat_exchangers.compute_Reynolds(
            double_u_tube_series.m_flow_pipe, r_in, epsilon, self.fluid)

        # Reynolds number
        self.assertAlmostEqual(Re, 11667.361867903752, places=8)
        # Pipe resistance (K/(W/m))
        self.assertAlmostEqual(double_u_tube_series.R_p, 0.08389296717975074, places=8)
        # Convection coefficient (W/(m2.K))
        self.assertAlmostEqual(double_u_tube_series.h_f, 2522.2943235234375, places=8)
        # Convective resistance (K/(W/m))
        self.assertAlmostEqual(
            double_u_tube_series.R_fp, 0.08973549264054603, places=8)
        # Borehole thermal resistance ((m.K)/W)
        self.assertAlmostEqual(
            double_u_tube_series.compute_effective_borehole_resistance(),
            0.16239288378526318, places=8)

    def test_coaxial(self):
        # Pipe dimensions
        # Inner pipe radii
        r_in_in = 44.2 / 1000. / 2.
        r_in_out = 50. / 1000. / 2.
        # Outer pipe radii
        r_out_in = 97.4 / 1000. / 2.
        r_out_out = 110. / 1000. / 2.
        # Pipe radii
        # Note: This convention is different from pygfunction
        r_inner = [r_in_in,
                   r_in_out]  # The radii of the inner pipe from in to out
        r_outer = [r_out_in,
                   r_out_out]  # The radii of the outer pipe from in to out
        epsilon = 1.0e-6  # Pipe roughness (m)

        # Pipe positioning
        pos = (0, 0)
        s = 0

        # Thermal properties
        k_p = [0.4, 0.4]  # Inner and outer pipe thermal conductivity (W/m.K)

        # Volumetric heat capacities
        rhoCp_p = 1542. * 1000.  # Pipe volumetric heat capacity (J/K.m3)

        # Thermal properties
        # Pipe
        pipe = plat.media.Pipe(pos, r_inner, r_outer, s, epsilon, k_p, rhoCp_p)

        # Define a borehole
        borehole = gt.boreholes.Borehole(self.H, self.D, self.r_b, x=0., y=0.)

        coaxial = plat.borehole_heat_exchangers.CoaxialPipe(
            self.m_flow_borehole, self.fluid, borehole, pipe, self.grout,
            self.soil)

        Re = plat.borehole_heat_exchangers.compute_Reynolds(
            coaxial.m_flow_pipe, coaxial.pipe.r_out[1], epsilon, self.fluid)

        # Reynolds number
        self.assertAlmostEqual(Re, 2291.0456031520093, places=8)
        # Convection coefficient (W/(m2.K))
        self.assertAlmostEqual(coaxial.h_fluid_a_out, 57.129718138096315, places=8)
        # Convective resistance (K/(W/m))
        self.assertAlmostEqual(
            coaxial.R_fp, 0.10560900480844831, places=8)
        # Borehole thermal resistance ((m.K)/W)
        self.assertAlmostEqual(
            coaxial.compute_effective_borehole_resistance(),
            0.19286205271006363, places=8)

    def test_coaxial_to_single_u_tube(self):
        # Pipe dimensions
        # Inner pipe radii
        r_in_in = 44.2 / 1000. / 2.
        r_in_out = 50. / 1000. / 2.
        # Outer pipe radii
        r_out_in = 97.4 / 1000. / 2.
        r_out_out = 110. / 1000. / 2.
        # Pipe radii
        # Note: This convention is different from pygfunction
        r_inner = [r_in_in,
                   r_in_out]  # The radii of the inner pipe from in to out
        r_outer = [r_out_in,
                   r_out_out]  # The radii of the outer pipe from in to out
        epsilon = 1.0e-6  # Pipe roughness (m)

        # Pipe positioning
        pos = (0, 0)
        s = 0

        # Thermal properties
        k_p = [0.4, 0.4]  # Inner and outer pipe thermal conductivity (W/m.K)

        # Thermal properties
        # Pipe
        pipe = plat.media.Pipe(
            pos, r_inner, r_outer, s, epsilon, k_p, self.rhoCp_p)

        # Define a borehole
        borehole = gt.boreholes.Borehole(self.H, self.D, self.r_b, x=0., y=0.)

        coaxial = plat.borehole_heat_exchangers.CoaxialPipe(
            self.m_flow_borehole, self.fluid, borehole, pipe, self.grout,
            self.soil)

        # Test intermediate variables
        V_fluid, V_pipe, R_conv, R_pipe = \
            plat.equivalance.concentric_tube_volumes(coaxial)
        self.assertAlmostEqual(V_fluid, 0.007021773740038546, places=8)
        self.assertAlmostEqual(V_pipe, 0.0024815440370705762, places=8)
        self.assertAlmostEqual(R_conv, 0.04531412438364096, places=8)
        self.assertAlmostEqual(R_pipe, 0.048404650347060686, places=8)

        single_u_tube = plat.equivalance.compute_equivalent(coaxial)

        # Test the single U-tube equivalent parameters
        V_flow = single_u_tube.m_flow_pipe * 1000. / single_u_tube.fluid.rho
        self.assertAlmostEqual(V_flow, 0.2)
        self.assertAlmostEqual(single_u_tube.r_in, 0.03342977714553299)
        self.assertAlmostEqual(single_u_tube.r_out, 0.03889087296526011)
        self.assertAlmostEqual(single_u_tube.pipe.s, 0.005370899457402695)
        self.assertAlmostEqual(single_u_tube.h_f, 247.4249791690743)
        self.assertAlmostEqual(single_u_tube.pipe.k, 0.3233497585764886)
        self.assertAlmostEqual(single_u_tube.grout.k, 0.5354034529364631)
        self.assertAlmostEqual(
            single_u_tube.compute_effective_borehole_resistance(),
            0.19286205242585608)

    def test_double_to_single_u_tube(self):
        # Pipe dimensions
        r_out = 26.67 / 1000. / 2.  # Pipe outer radius (m)
        r_in = 21.6 / 1000. / 2.  # Pipe inner radius (m)
        s = 32.3 / 1000.  # Inner-tube to inner-tube Shank spacing (m)
        epsilon = 1.0e-6  # Pipe roughness (m)

        # Pipe positions
        # Double U-tube [(x_in, y_in), (x_out, y_out), (x_in, y_in), (x_out, y_out)]
        pos = plat.media.Pipe.place_pipes(s, r_out, 2)

        # Thermal properties
        # Pipe
        pipe = plat.media.Pipe(pos, r_in, r_out, s, epsilon, self.k_p, self.rhoCp_p)

        # Define a borehole
        borehole = gt.boreholes.Borehole(self.H, self.D, self.r_b, x=0., y=0.)

        # Double U-tube defaults to parallel
        double_u_tube = plat.borehole_heat_exchangers.MultipleUTube(
            self.m_flow_borehole, self.fluid, borehole, pipe, self.grout,
            self.soil)

        # Test intermediate variables
        V_fluid, V_pipe, R_conv, R_pipe = \
            plat.equivalance.u_tube_volumes(double_u_tube)
        self.assertAlmostEqual(V_fluid, 0.001465741468458854)
        self.assertAlmostEqual(V_pipe, 0.0007688385143611112)
        self.assertAlmostEqual(R_conv, 0.13198055664449054)
        self.assertAlmostEqual(R_pipe, 0.020973241794937685)

        single_u_tube = plat.equivalance.compute_equivalent(double_u_tube)

        # Test the single U-tube equivalent parameters
        V_flow = single_u_tube.m_flow_pipe * 1000. / single_u_tube.fluid.rho
        self.assertAlmostEqual(V_flow, 0.2)
        self.assertAlmostEqual(single_u_tube.r_in, 0.015273506473629427)
        self.assertAlmostEqual(single_u_tube.r_out, 0.01885853785424522)
        self.assertAlmostEqual(single_u_tube.pipe.s, 0.02485528286100637)
        self.assertAlmostEqual(single_u_tube.h_f, 1287.1059002600034)
        self.assertAlmostEqual(single_u_tube.pipe.k, 0.2316554011453535)
        self.assertAlmostEqual(single_u_tube.grout.k, 1.6966493055905383)
        self.assertAlmostEqual(
            single_u_tube.compute_effective_borehole_resistance(),
            0.15970357274051686)



import unittest
import os

import ghedt as dt
import ghedt.peak_load_analysis_tool as plat
import pygfunction as gt

import pandas as pd

TESTDATA_FILENAME = os.path.join(os.path.dirname(__file__),
                                 'Atlanta_Office_Building_Loads.csv')


class TestGHE(unittest.TestCase):

    def setUp(self) -> None:
        # Borehole dimensions
        # -------------------
        self.H = 100.  # Borehole length (m)
        self.D = 2.  # Borehole buried depth (m)
        self.r_b = 0.075  # Borehole radius]
        self.B = 5.  # Borehole spacing (m)

        # Pipe dimensions
        # ---------------
        # U-tubes
        r_out = 26.67 / 1000. / 2.  # Pipe outer radius (m)
        r_in = 21.6 / 1000. / 2.  # Pipe inner radius (m)
        s = 32.3 / 1000.  # Inner-tube to inner-tube Shank spacing (m)
        # Coaxial
        # Inner pipe radii
        r_in_in = 44.2 / 1000. / 2.
        r_in_out = 50. / 1000. / 2.
        # Outer pipe radii
        r_out_in = 97.4 / 1000. / 2.
        r_out_out = 110. / 1000. / 2.
        # Pipe radii
        # Note: This convention is different from pygfunction
        r_inner = [r_in_in,
                   r_in_out]  # The radii of the inner pipe from in to out
        r_outer = [r_out_in,
                   r_out_out]  # The radii of the outer pipe from in to out

        epsilon = 1.0e-6  # Pipe roughness (m)

        # Pipe positions
        # --------------
        # Single U-tube [(x_in, y_in), (x_out, y_out)]
        pos_s = plat.media.Pipe.place_pipes(s, r_out, 1)
        # Double U-tube
        pos_d = plat.media.Pipe.place_pipes(s, r_out, 2)
        # Coaxial
        pos_c = (0, 0)

        # Thermal conductivities
        # ----------------------
        k_p = 0.4  # Pipe thermal conductivity (W/m.K)
        k_s = 2.0  # Ground thermal conductivity (W/m.K)
        k_g = 1.0  # Grout thermal conductivity (W/m.K)
        # Pipe thermal conductivity list for coaxial
        k_p_c = [0.4, 0.4]  # Inner and outer pipe thermal conductivity (W/m.K)

        # Volumetric heat capacities
        # --------------------------
        rhoCp_p = 1542. * 1000.  # Pipe volumetric heat capacity (J/K.m3)
        rhoCp_s = 2343.493 * 1000.  # Soil volumetric heat capacity (J/K.m3)
        rhoCp_g = 3901. * 1000.  # Grout volumetric heat capacity (J/K.m3)

        # Thermal properties
        # ------------------
        # Pipe
        self.pipe_s = \
            plat.media.Pipe(pos_s, r_in, r_out, s, epsilon, k_p, rhoCp_p)
        self.pipe_d = \
            plat.media.Pipe(pos_d, r_in, r_out, s, epsilon, k_p, rhoCp_p)
        self.pipe_c = \
            plat.media.Pipe(pos_c, r_inner, r_outer, s, epsilon, k_p_c, rhoCp_p)

        # Single U-tube BHE object
        self.SingleUTube = plat.borehole_heat_exchangers.SingleUTube
        # Double U-tube bhe object
        self.DoubleUTube = plat.borehole_heat_exchangers.MultipleUTube
        # Coaxial tube bhe object
        self.CoaxialTube = plat.borehole_heat_exchangers.CoaxialPipe

        # Soil
        ugt = 18.3  # Undisturbed ground temperature (degrees Celsius)
        self.soil = plat.media.Soil(k_s, rhoCp_s, ugt)
        # Grout
        self.grout = plat.media.Grout(k_g, rhoCp_g)

        # Coordinates
        Nx = 12
        Ny = 13
        self.coordinates = dt.coordinates.rectangle(Nx, Ny, self.B, self.B)

        # Compute a range of g-functions for interpolation
        self.log_time = dt.utilities.Eskilson_log_times()
        self.H_values = [24., 48., 96., 192., 384.]
        self.r_b_values = [self.r_b] * len(self.H_values)
        self.D_values = [2.] * len(self.H_values)

        # Inputs related to fluid
        # -----------------------
        V_flow_borehole = 0.2  # System volumetric flow rate (L/s)
        mixer = 'MEG'  # Ethylene glycol mixed with water
        percent = 0.  # Percentage of ethylene glycol added in

        # -----------------------
        # Fluid properties
        self.fluid = gt.media.Fluid(mixer, percent)
        self.V_flow_system = V_flow_borehole * float(
            Nx * Ny)  # System volumetric flow rate (L/s)
        # Total fluid mass flow rate per borehole (kg/s)
        self.m_flow_borehole = V_flow_borehole / 1000. * self.fluid.rho

        # Simulation start month and end month
        # --------------------------------
        # Simulation start month and end month
        start_month = 1
        n_years = 20
        end_month = n_years * 12
        # Maximum and minimum allowable fluid temperatures
        max_EFT_allowable = 35  # degrees Celsius
        min_EFT_allowable = 5  # degrees Celsius
        # Maximum and minimum allowable heights
        max_Height = 384  # in meters
        min_Height = 24  # in meters
        self.sim_params = plat.media.SimulationParameters(
            start_month, end_month, max_EFT_allowable, min_EFT_allowable,
            max_Height, min_Height)

        # Process loads from file
        # -----------------------
        # read in the csv file and convert the loads to a list of length 8760
        hourly_extraction: dict = \
            pd.read_csv(TESTDATA_FILENAME).to_dict('list')
        # Take only the first column in the dictionary
        self.hourly_extraction_ground_loads: list = \
            hourly_extraction[list(hourly_extraction.keys())[0]]

    def test_single_u_tube(self):

        # Define a borehole
        borehole = gt.boreholes.Borehole(self.H, self.D, self.r_b, x=0., y=0.)

        # Initialize GHE object
        g_function = dt.gfunction.compute_live_g_function(
            self.B, self.H_values, self.r_b_values, self.D_values,
            self.m_flow_borehole, self.SingleUTube,
            self.log_time, self.coordinates, self.fluid, self.pipe_s,
            self.grout, self.soil)

        # Initialize the GHE object
        ghe = dt.ground_heat_exchangers.GHE(
            self.V_flow_system, self.B, self.SingleUTube, self.fluid,
            borehole, self.pipe_s, self.grout, self.soil,
            g_function, self.sim_params, self.hourly_extraction_ground_loads)

        max_HP_EFT, min_HP_EFT = ghe.simulate(method='hybrid')

        self.assertAlmostEqual(39.1133223639883, max_HP_EFT)
        self.assertAlmostEqual(16.657425746458046, min_HP_EFT)

        ghe.size(method='hybrid')

        self.assertAlmostEqual(ghe.bhe.b.H, 130.34501297366248, places=2)

    def test_double_u_tube(self):

        # Define a borehole
        borehole = gt.boreholes.Borehole(self.H, self.D, self.r_b, x=0., y=0.)

        # Initialize GHE object
        g_function = dt.gfunction.compute_live_g_function(
            self.B, self.H_values, self.r_b_values, self.D_values,
            self.m_flow_borehole, self.DoubleUTube,
            self.log_time, self.coordinates, self.fluid, self.pipe_d,
            self.grout, self.soil)

        # Initialize the GHE object
        ghe = dt.ground_heat_exchangers.GHE(
            self.V_flow_system, self.B, self.DoubleUTube, self.fluid,
            borehole, self.pipe_d, self.grout, self.soil,
            g_function, self.sim_params, self.hourly_extraction_ground_loads)

        max_HP_EFT, min_HP_EFT = ghe.simulate(method='hybrid')

        self.assertAlmostEqual(38.001534870972144, max_HP_EFT, places=6)
        self.assertAlmostEqual(16.985639409819775, min_HP_EFT, places=6)

        ghe.size(method='hybrid')

        self.assertAlmostEqual(ghe.bhe.b.H, 122.111390772492, places=2)

    def test_coaxial_tube(self):

        # Define a borehole
        borehole = gt.boreholes.Borehole(self.H, self.D, self.r_b, x=0., y=0.)

        # Initialize GHE object
        g_function = dt.gfunction.compute_live_g_function(
            self.B, self.H_values, self.r_b_values, self.D_values,
            self.m_flow_borehole, self.CoaxialTube,
            self.log_time, self.coordinates, self.fluid, self.pipe_c,
            self.grout, self.soil)

        # Re-Initialize the GHE object
        ghe = dt.ground_heat_exchangers.GHE(
            self.V_flow_system, self.B, self.CoaxialTube, self.fluid,
            borehole, self.pipe_c, self.grout, self.soil,
            g_function, self.sim_params, self.hourly_extraction_ground_loads)

        max_HP_EFT, min_HP_EFT = ghe.simulate(method='hybrid')

        self.assertAlmostEqual(37.77089369190098, max_HP_EFT)
        self.assertAlmostEqual(17.61071203099197, min_HP_EFT)

        ghe.size(method='hybrid')

        self.assertAlmostEqual(ghe.bhe.b.H, 121.244590801587, places=2)



import numpy as np


class ThermalProperty:
    def __init__(self, k, rhoCp):
        self.k = k          # Thermal conductivity (W/m.K)
        self.rhoCp = rhoCp  # Volumetric heat capacity (J/K.m3)

    def __repr__(self):
        def justify(category, value):
            return category.ljust(40) + '= ' + value + '\n'

        output = str(self.__class__) + '\n'

        if type(self.k) == float:
            output += justify('Thermal conductivity',
                              str(round(self.k, 4)) + ' (W/m.K)')
        else:
            output += justify('Thermal conductivity',
                              str(self.k) + ' (W/m.K)')
        output += justify('Volumetric heat capacity',
                          str(round(self.rhoCp, 4)) + ' (J/K.m3)')

        return output


class Grout(ThermalProperty):
    def __init__(self, k, rhoCp):
        # Make variables from ThermalProperty available to Grout
        ThermalProperty.__init__(self, k, rhoCp)


class Pipe(ThermalProperty):
    def __init__(self, pos, r_in, r_out, s, eps, k, rhoCp):
        # Make variables from ThermalProperty available to Pipe
        ThermalProperty.__init__(self, k, rhoCp)

        # Pipe specific parameters
        self.pos = pos      # Pipe positions either a list of tuples or tuple
        self.r_in = r_in    # Pipe inner radius (m) can be float or list
        self.r_out = r_out  # Pipe outer radius (m) can be float or list
        self.s = s          # Center pipe to center pipe shank spacing
        self.eps = eps      # Pipe roughness (m)
        if type(pos) is list:
            self.n_pipes = int(len(pos) / 2)  # Number of pipes
        else:
            self.n_pipes = 1

    def __repr__(self):
        def justify(category, value):
            return category.ljust(40) + '= ' + value + '\n'

        output = ThermalProperty.__repr__(self)

        output += justify('Pipe Positions (Center of pipes)', str(self.pos))
        if type(self.r_in) == float:
            output += justify('Pipe inner radius',
                              str(round(self.r_in, 4)) + ' (m)')
            output += justify('Pipe outer radius',
                              str(round(self.r_out, 4)) + ' (m)')
        else:
            output += justify('Pipe inner radii',
                              str(self.r_in) + ' (m)')
            output += justify('Pipe outer radii',
                              str(self.r_out) + ' (m)')
        output += justify('Shank spacing (pipe to pipe)',
                          str(round(self.s, 4)) + ' (m)')
        output += justify('Pipe roughness', str(self.eps) + ' (m)')
        output += justify('Number of pipes', str(self.n_pipes))

        return output

    @staticmethod
    def place_pipes(s, r_out, n_pipes):
        """ Positions pipes in an axisymetric configuration."""
        D_s = s / 2 + r_out
        pi = np.pi
        dt = pi / float(n_pipes)
        pos = [(0., 0.) for i in range(2 * n_pipes)]
        for i in range(n_pipes):
            pos[2 * i] = (
                D_s * np.cos(2.0 * i * dt + pi),
                D_s * np.sin(2.0 * i * dt + pi))
            pos[2 * i + 1] = (D_s * np.cos(2.0 * i * dt + pi + dt),
                              D_s * np.sin(2.0 * i * dt + pi + dt))
        return pos


class Soil(ThermalProperty):
    def __init__(self, k, rhoCp, ugt):
        # Make variables from ThermalProperty available to Pipe
        ThermalProperty.__init__(self, k, rhoCp)

        # Soil specific parameters
        self.ugt = ugt

    def __repr__(self):
        def justify(category, value):
            return category.ljust(40) + '= ' + value + '\n'

        output = ThermalProperty.__repr__(self)

        output += justify('Undisturbed ground temperature',
                          str(round(self.ugt, 4)) + ' (degrees Celsius)')

        return output


class SimulationParameters:
    def __init__(self, start_month, end_month, max_EFT_allowable,
                 min_EFT_allowable, max_Height, min_Height):
        # Simulation parameters not found in other objects
        # ------------------------------------------------
        # Simulation start month and end month
        self.start_month = start_month
        self.end_month = end_month
        # Maximum and minimum allowable fluid temperatures
        self.max_EFT_allowable = max_EFT_allowable  # degrees Celsius
        self.min_EFT_allowable = min_EFT_allowable  # degrees Celsius
        # Maximum and minimum allowable heights
        self.max_Height = max_Height  # in meters
        self.min_Height = min_Height  # in meters

    def __repr__(self):
        def justify(category, value):
            return category.ljust(40) + '= ' + value + '\n'

        output = str(self.__class__) + '\n'

        output += justify('Start month', str(self.start_month) + ' (months)')
        output += justify('End month', str(self.end_month) + '(months)')
        output += justify('Max EFT Allowable',
                          str(self.max_EFT_allowable) + ' (degrees Celsius)')
        output += justify('Min EFT Allowable',
                          str(self.min_EFT_allowable) + ' (degrees Celsius)')
        output += justify('Maximum height', str(self.max_Height) + ' (m)')
        output += justify('Minimum height', str(self.min_Height) + ' (m)')

        return output



import copy
import warnings

import scipy.interpolate
import scipy.optimize
import pygfunction as gt
import ghedt.peak_load_analysis_tool as plat
import ghedt as dt

import numpy as np


class BaseGHE:
    def __init__(
            self, V_flow_system: float, B_spacing: float,
            bhe_function: plat.borehole_heat_exchangers,
            fluid: gt.media.Fluid, borehole: gt.boreholes.Borehole,
            pipe: plat.media.Pipe, grout: plat.media.Grout,
            soil: plat.media.Soil, GFunction: dt.gfunction.GFunction,
            sim_params: plat.media.SimulationParameters,
            hourly_extraction_ground_loads: list):

        self.V_flow_system = V_flow_system
        self.B_spacing = B_spacing
        self.nbh = float(len(GFunction.bore_locations))
        self.V_flow_borehole = self.V_flow_system / self.nbh
        m_flow_borehole = self.V_flow_borehole / 1000. * fluid.rho

        # Borehole Heat Exchanger
        self.bhe_object = bhe_function
        self.bhe = bhe_function(
            m_flow_borehole, fluid, borehole, pipe, grout, soil)
        # Equivalent borehole Heat Exchanger
        self.bhe_eq = plat.equivalance.compute_equivalent(self.bhe)

        # Radial numerical short time step
        self.radial_numerical = \
            plat.radial_numerical_borehole.RadialNumericalBH(self.bhe_eq)
        self.radial_numerical.calc_sts_g_functions(self.bhe_eq)

        # GFunction object
        self.GFunction = GFunction
        # Additional simulation parameters
        self.sim_params = sim_params
        # Hourly ground extraction loads
        # Building cooling is negative, building heating is positive
        self.hourly_extraction_ground_loads = hourly_extraction_ground_loads

    @staticmethod
    def header(text):
        return 50 * '-' + '\n' + '|' + text.center(48) + \
               '|\n' + 50 * '-' + '\n'

    @staticmethod
    def justify(category, value):
        return category.ljust(40) + '= ' + value + '\n'

    def __repr__(self):
        header = self.header
        # Header
        output = 50 * '-' + '\n'
        output += header('GHEDT GHE Output - Version 0.1')
        output += 50 * '-' + '\n'

        def justify(category, value):
            return category.ljust(40) + '= ' + value + '\n'

        # Detailed information
        output += justify('Number of borheoles',
                          str(len(self.GFunction.bore_locations)))
        output += justify('Depth of the borehole',
                          str(round(self.bhe.b.H, 4)) + ' (m)')
        output += justify('Bore hole spacing',
                          str(round(self.B_spacing, 4)) + ' (m)')
        output += header('Borehole Heat Exchanger')
        output += self.bhe.__repr__()
        output += header('Equivalent Borehole Heat Exchanger')
        output += self.bhe_eq.__repr__()
        output += header('Simulation parameters')
        output += self.sim_params.__repr__()

        return output

    @staticmethod
    def combine_sts_lts(log_time_lts: list, g_lts: list, log_time_sts: list,
                        g_sts: list) -> scipy.interpolate.interp1d:
        # make sure the short time step doesn't overlap with the long time step
        max_log_time_sts = max(log_time_sts)
        min_log_time_lts = min(log_time_lts)

        if max_log_time_sts < min_log_time_lts:
            log_time = log_time_sts + log_time_lts
            g = g_sts + g_lts
        else:
            # find where to stop in sts
            i = 0
            value = log_time_sts[i]
            while value <= min_log_time_lts:
                i += 1
                value = log_time_sts[i]
            log_time = log_time_sts[0:i] + log_time_lts
            g = g_sts[0:i] + g_lts

        g = scipy.interpolate.interp1d(log_time, g)

        return g

    def grab_g_function(self, B_over_H):
        # interpolate for the Long time step g-function
        g_function, rb_value, D_value, H_eq = \
            self.GFunction.g_function_interpolation(B_over_H)
        # correct the long time step for borehole radius
        g_function_corrected = \
            self.GFunction.borehole_radius_correction(g_function,
                                                      rb_value,
                                                      self.bhe.b.r_b)
        # Don't Update the HybridLoad (its dependent on the STS) because
        # it doesn't change the results much and it slows things down a lot
        # combine the short and long time step g-function
        g = self.combine_sts_lts(
            self.GFunction.log_time, g_function_corrected,
            self.radial_numerical.lntts.tolist(),
            self.radial_numerical.g.tolist())

        return g

    def cost(self, max_EFT, min_EFT):
        delta_T_max = max_EFT - self.sim_params.max_EFT_allowable
        delta_T_min = self.sim_params.min_EFT_allowable - min_EFT

        T_excess = max([delta_T_max, delta_T_min])

        return T_excess

    def _simulate_detailed(self, Q_dot: np.ndarray, time_values: np.ndarray,
                           g: scipy.interpolate.interp1d):
        # Perform a detailed simulation based on a numpy array of heat rejection
        # rates, Q_dot (Watts) where each load is applied at the time_value
        # (seconds). The g-function can interpolated.
        # Source: Chapter 2 of Advances in Ground Source Heat Pumps

        n = Q_dot.size

        # Convert the total load applied to the field to the average over
        # borehole wall rejection rate
        # At time t=0, make the heat rejection rate 0.
        Q_dot_b = np.hstack((0., Q_dot / float(self.nbh)))
        time_values = np.hstack((0., time_values))

        Q_dot_b_dt = np.hstack((Q_dot_b[1:] - Q_dot_b[:-1]))

        ts = self.radial_numerical.t_s  # (-)
        two_pi_k = 2. * np.pi * self.bhe.soil.k  # (W/m.K)
        H = self.bhe.b.H  # (meters)
        Tg = self.bhe.soil.ugt  # (Celsius)
        Rb = self.bhe.compute_effective_borehole_resistance()  # (m.K/W)
        m_dot = self.bhe.m_flow_borehole  # (kg/s)
        cp = self.bhe.fluid.cp  # (J/kg.s)

        HPEFT = []
        delta_Tb = []
        for i in range(1, n+1):
            # Take the last i elements of the reversed time array
            _time = time_values[i] - time_values[0:i]
            # _time = time_values_reversed[n - i:n]
            g_values = g(np.log((_time * 3600.) / ts))
            # Tb = Tg + (q_dt * g)  (Equation 2.12)
            delta_Tb_i = (Q_dot_b_dt[0:i] / H / two_pi_k).dot(g_values)
            # Tf = Tb + q_i * R_b^* (Equation 2.13)
            Tb = Tg + delta_Tb_i
            # Bulk fluid temperature
            Tf_bulk = Tb + Q_dot_b[i] / H * Rb
            # T_out = T_f - Q / (2 * mdot cp)  (Equation 2.14)
            Tf_out = Tf_bulk - Q_dot_b[i] / (2 * m_dot * cp)
            HPEFT.append(Tf_out)
            delta_Tb.append(delta_Tb_i)

        return HPEFT, delta_Tb

    def compute_g_functions(self):
        # Compute g-functions for a bracketed solution, based on min and max
        # height
        min_height = self.sim_params.min_Height
        max_height = self.sim_params.max_Height
        avg_height = (min_height + max_height) / 2.
        H_values = [min_height, avg_height, max_height]
        r_b_values = [self.bhe.b.r_b] * len(H_values)
        D_values = [self.bhe.b.D] * len(H_values)

        coordinates = self.GFunction.bore_locations
        log_time = self.GFunction.log_time

        g_function = dt.gfunction.compute_live_g_function(
            self.B_spacing, H_values, r_b_values, D_values,
            self.bhe.m_flow_borehole, self.bhe_object, log_time,
            coordinates, self.bhe.fluid, self.bhe.pipe,
            self.bhe.grout, self.bhe.soil)

        self.GFunction = g_function

        return


class GHE(BaseGHE):
    def __init__(self, V_flow_system: float, B_spacing: float,
                 bhe_object: plat.borehole_heat_exchangers,
                 fluid: gt.media.Fluid, borehole: gt.boreholes.Borehole,
                 pipe: plat.media.Pipe, grout: plat.media.Grout,
                 soil: plat.media.Soil,
                 GFunction: dt.gfunction.GFunction,
                 sim_params: plat.media.SimulationParameters,
                 hourly_extraction_ground_loads: list
                 ):
        BaseGHE.__init__(
            self, V_flow_system, B_spacing, bhe_object, fluid, borehole, pipe,
            grout, soil, GFunction, sim_params, hourly_extraction_ground_loads)

        # Split the extraction loads into heating and cooling for input to
        # the HybridLoad object
        hourly_rejection_loads, hourly_extraction_loads = \
            plat.ground_loads.HybridLoad.split_heat_and_cool(
                self.hourly_extraction_ground_loads)

        hybrid_load = plat.ground_loads.HybridLoad(
            hourly_rejection_loads, hourly_extraction_loads, self.bhe_eq,
            self.radial_numerical, sim_params)

        # hybrid load object
        self.hybrid_load = hybrid_load

        # List of heat pump exiting fluid temperatures
        self.HPEFT = []
        # list of change in borehole wall temperatures
        self.dTb = []

    def __repr__(self):
        output = BaseGHE.__repr__(self)

        self.header('Simulation Results')
        if len(self.HPEFT) > 0:
            max_HP_EFT = max(self.HPEFT)
            min_HP_EFT = min(self.HPEFT)
            output += self.justify('Max HP entering temp',
                                   str(round(max_HP_EFT, 4)) +
                                   ' (degrees Celsius)')
            output += self.justify('Min HP entering temp',
                                   str(round(min_HP_EFT, 4)) +
                                   ' (degrees Celsius)')
            T_excess = self.cost(max_HP_EFT, min_HP_EFT)
            output += self.justify('Excess fluid temperature',
                                   str(round(T_excess, 4)) +
                                   ' (degrees Celsius)')
        output += self.header('Peak Load Analysis')
        output += self.hybrid_load.__repr__() + '\n'
        output += self.header('GFunction Information')
        output += 'Coordinates\nx(m)\ty(m)\n'
        for i in range(len(self.GFunction.bore_locations)):
            x, y = self.GFunction.bore_locations[i]
            output += str(x) + '\t' + str(y) + '\n'
        output += 'G-Function\nln(t/ts)\tg\n'
        B_over_H = self.B_spacing / self.bhe.b.H
        g = self.grab_g_function(B_over_H)
        total_g_values = g.x.size
        number_lts_g_values = 27
        number_sts_g_values = 50
        sts_step_size = int(np.floor((total_g_values - number_lts_g_values) /
                                     number_sts_g_values).tolist())
        lntts = []
        g_values = []
        for i in range(0, (total_g_values - number_lts_g_values),
                       sts_step_size):
            lntts.append(g.x[i].tolist())
            g_values.append(g.y[i].tolist())
        lntts += g.x[
                 total_g_values - number_lts_g_values: total_g_values].tolist()
        g_values += g.y[
                    total_g_values - number_lts_g_values: total_g_values].tolist()
        for i in range(len(lntts)):
            output += str(round(lntts[i], 4)) + '\t' + \
                      str(round(g_values[i], 4)) + '\n'
        return output

    def simulate(self, method='hybrid'):
        B = self.B_spacing
        B_over_H = B / self.bhe.b.H

        self.bhe.update_thermal_resistance()

        # Solve for equivalent single U-tube
        self.bhe_eq = plat.equivalance.compute_equivalent(self.bhe)
        # Update short time step object with equivalent single u-tube
        self.radial_numerical.calc_sts_g_functions(self.bhe_eq)
        # Combine the short and long-term g-functions. The long term g-function
        # is interpolated for specific B/H and rb/H values.
        g = self.grab_g_function(B_over_H)

        if method == 'hybrid':
            Q_dot = self.hybrid_load.load[2:] * 1000.  # convert to Watts
            time_values = self.hybrid_load.hour[2:]  # convert to seconds

            HPEFT, dTb = self._simulate_detailed(Q_dot, time_values, g)
        elif method == 'hourly':
            n_months = \
                self.sim_params.end_month - self.sim_params.start_month + 1
            n_hours = int(n_months / 12. * 8760.)
            Q_dot = copy.deepcopy(self.hourly_extraction_ground_loads)
            # How many times does q need to be repeated?
            n_years = int(np.ceil(n_hours / 8760))
            Q_dot = Q_dot * n_years
            Q_dot = -1. * np.array(Q_dot)  # Convert loads to rejection
            t = np.arange(1, n_hours + 1, 1)

            HPEFT, dTb = self._simulate_detailed(Q_dot, t, g)
        else:
            raise ValueError('Only hybrid or hourly methods available.')

        self.HPEFT = HPEFT
        self.dTb = dTb

        max_HP_EFT = float(max(HPEFT))
        min_HP_EFT = float(min(HPEFT))
        return max_HP_EFT, min_HP_EFT

    def size(self, method='hybrid') -> None:
        # Size the ground heat exchanger

        def local_objective(H):
            self.bhe.b.H = H
            max_HP_EFT, min_HP_EFT = self.simulate(method=method)
            T_excess = self.cost(max_HP_EFT, min_HP_EFT)
            return T_excess

        # Make the initial guess variable the average of the heights given
        self.bhe.b.H = \
            (self.sim_params.max_Height + self.sim_params.min_Height) / 2.
        # bhe.b.H is updated during sizing
        plat.equivalance.solve_root(
            self.bhe.b.H, local_objective, lower=self.sim_params.min_Height,
            upper=self.sim_params.max_Height, xtol=1.0e-6, rtol=1.0e-6,
            maxiter=50)
        if self.bhe.b.H == self.sim_params.min_Height:
            warnings.warn('The minimum height provided to size this ground heat'
                          ' exchanger is not shallow enough. Provide a '
                          'shallower allowable depth or decrease the size of '
                          'the heat exchanger.')
        if self.bhe.b.H == self.sim_params.max_Height:
            warnings.warn('The maximum height provided to size this ground '
                          'heat exchanger is not deep enough. Provide a deeper '
                          'allowable depth or increase the size of the heat '
                          'exchanger.')

        return



from setuptools import setup, find_packages

from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='ghedt',
      install_requires=['pygfunction>=2.1',
                        'pandas>=1.3.2',
                        'openpyxl>=3.0.8',
                        'opencv-python==4.5.4.58'],
      url='https://github.com/j-c-cook/ghedt',
      download_url='https://github.com/j-c-cook/ghedt/archive/v0.3.1.tar.gz',
      long_description=long_description,
      long_description_content_type='text/markdown',
      version='0.3.1',
      packages=find_packages(),
      include_package_data=True,
      author='Jack C. Cook',
      author_email='jack-c-cook@protonmail.com',
      description='A ground heat exchanger design tool with the advanced and '
                  'unmatched capability of automatic borehole field selection '
                  'based on drilling geometric land constraints.')
